@@ -18,6 +18,7 @@ differences between two configurations.
 
 
 from copy import deepcopy
+from itertools import chain
 import re
 import sys
 
@@ -104,11 +105,8 @@ class DiffConvert:
     this is useful when a higher level in the configuration dictionary
     is required in the action function
 
-    name -- if there is more than one action acting at a particular path
-    point, this can be defined to differentiate them: it's useful when
-    using the insert_before parameter, but isn't needed if the order
-    does not matter (except for ease of debugging, to clarify which
-    converter is being called)
+    block -- the block of configuration in which this converter is to be
+    applied (see DiffConfig)
 
     insert_before -- if this rule needs to happen prior to the actions
     of another rule, this 2-tuple gives the (path, name) before the
@@ -122,8 +120,8 @@ class DiffConvert:
     context = tuple()
     cmd = None    # leave as None to cause an error if not defined in child
     ext = tuple()
-    name = None
-    insert_before = None
+    block = None
+    triggers = set()
 
 
     def __init__(self):
@@ -234,6 +232,21 @@ class DiffConvert:
 
         return self._path_matches(d, self._path_full)
 
+    def explicit_context_matches(self, d, context):
+        """This method is used when checking for trigger matches.
+
+        It is similar to full_matches() except the explicit context
+        (without wildcards) is supplied and the command path appended;
+        the matches are returned as a list.
+
+        The context will have been taken from the converter which set up
+        the trigger (with the wildcarded fields completed) so we only
+        want to match parts of the configuration which have that
+        explicit context.
+        """
+
+        return self._path_matches(d, tuple(context) + self.cmd)
+
 
     def context_removed(self, d, match):
         """This method takes a remove dictionary, as returned by
@@ -295,6 +308,74 @@ class DiffConvert:
         return True
 
 
+    def get_path(self):
+        """Get the non-extended path for this converter (i.e. the
+        'context' and 'cmd' paths joined, excluding the 'ext').
+        """
+
+        return self.context + self.cmd
+
+
+    def get_match_context(self, match):
+        """Return the context portion of a particular match for this
+        converter.
+
+        This is used when a trigger is set, to extract the context of a
+        converter.
+        """
+
+        return match[0 : self._context_len]
+
+
+    def get_cfg(self, cfg, match):
+        """This method returns the configuration to be passed to the
+        converter's action methods [remove() and update()].
+
+        By default, it indexes through the configuration using the path
+        in the converter.  Converters may wish to override this, in some
+        cases, for performance (perhaps if the entire configuration is
+        to be returned).
+
+        If the specified match path is not found, or there was a problem
+        indexing along the path, None is returned, rather than an
+        exception raised.
+        """
+
+        # try to get the specified matching path
+        try:
+            return deepget(cfg, *match[0 : self._path_len])
+
+        except TypeError:
+            # we got a TypeError so make the assumption that we've hit
+            # a non-indexable element (such as a set) as the final
+            # element of the path, so just return None
+            return None
+
+
+    def get_ext(self, cfg):
+        """This method gets the extension part of the path, given a
+        configuration dictionary starting at the path (i.e. what is
+        passed as 'old/new' and 'rem/upd' in the action methods).
+
+        An action method [remove() or update()] can use this to get the
+        extension portion without needing to explicitly index through
+        it.
+        """
+
+        return deepget(cfg, *self.ext)
+
+
+    def get_wildcard_args(self, match):
+        """This method returns a list of the wildcarded parts of the
+        specified match as a list.  For example, if the path of this
+        converter is 'interface', None) and the match is ['interface',
+        'Vlan100'], the returned result will be ['Vlan100'].  This is
+        used to supply the 'args' parameter to action methods.
+        """
+
+        return [ match[i] for i in self.wildcard_indices ]
+
+
     def remove(self, old, *args):
         """The remove() method is called when the specified path is in
         the remove differences (i.e. it's in the 'from' configuration
@@ -449,61 +530,18 @@ class DiffConvert:
         pass
 
 
-    def get_path(self):
-        """Get the non-extended path for this converter (i.e. the
-        'context' and 'cmd' paths joined, excluding the 'ext').
+    def trigger(self, old, upd, new, *args):
+        """The trigger() method is the same as the update() method
+        except it is called when a trigger match occurs.
+
+        By default, this method just calls the update() method but can
+        be overridden by subclasses.
+
+        Note that a trigger() method will not be called if a remove or
+        update match has already been called on a particular match.
         """
 
-        return self.context + self.cmd
-
-
-    def get_cfg(self, cfg, match):
-        """This method returns the configuration to be passed to the
-        converter's action methods [remove() and update()].
-
-        By default, it indexes through the configuration using the path
-        in the converter.  Converters may wish to override this, in some
-        cases, for performance (perhaps if the entire configuration is
-        to be returned).
-
-        If the specified match path is not found, or there was a problem
-        indexing along the path, None is returned, rather than an
-        exception raised.
-        """
-
-        # try to get the specified matching path
-        try:
-            return deepget(cfg, *match[0 : self._path_len])
-
-        except TypeError:
-            # we got a TypeError so make the assumption that we've hit
-            # a non-indexable element (such as a set) as the final
-            # element of the path, so just return None
-            return None
-
-
-    def get_ext(self, cfg):
-        """This method gets the extension part of the path, given a
-        configuration dictionary starting at the path (i.e. what is
-        passed as 'old/new' and 'rem/upd' in the action methods).
-
-        An action method [remove() or update()] can use this to get the
-        extension portion without needing to explicitly index through
-        it.
-        """
-
-        return deepget(cfg, *self.ext)
-
-
-    def get_wildcard_args(self, match):
-        """This method returns a list of the wildcarded parts of the
-        specified match as a list.  For example, if the path of this
-        converter is 'interface', None) and the match is ['interface',
-        'Vlan100'], the returned result will be ['Vlan100'].  This is
-        used to supply the 'args' parameter to action methods.
-        """
-
-        return [ match[i] for i in self.wildcard_indices ]
+        return self.update(old, upd, new, *args)
 
 
 
@@ -514,6 +552,13 @@ class DiffConfig:
     configuration files).
 
     It encapsulates the rules for exluding items from the comparison.
+
+    The list of converters is organised into 'blocks', each with a list
+    of converters.  The sequence of blocks is set in _init_blocks() and
+    each converter has an optional 'block' attribute which specifies
+    which block it is to be added to.  The converters are added with
+    _add_converters().  Both of these methods will likely need to be
+    overridden in concrete classes.
     """
 
 
@@ -522,7 +567,8 @@ class DiffConfig:
                  init_subtree_dump_filter=[]):
 
         """The constructor initialises the exclude list to empty and
-        adds and converters using _add_converters().  It also stores
+        adds the converter block sequence with _add_blocks() and
+        individual converters using _add_converters().  It also stores
         some settings controlling the level of information describing
         the conversion process, based on the command line arguments:
 
@@ -552,9 +598,32 @@ class DiffConfig:
         # to deepremoveitems()
         self.init_rules()
 
-        # initialise the list of converters and add them
-        self._cvts = []
+        # initialise the converter block sequence
+        self._init_blocks()
+
+        # the converters are stored in a dictionary, keyed on the block
+        # name, with each value a list of converters in that block, to
+        # be applied in the specified order
+        self._cvts = {}
+        for b in self._blocks:
+            self._cvts[b] = []
+
+        # add and sort the converters
         self._add_converters()
+        self._sort_converters()
+
+
+    def _add_blocks(self):
+        """This method adds the sequence of converter blocks.  Blocks
+        group together converters which must be run before others (in
+        other blocks).
+
+        The base class assumes a single block with no name ('None'),
+        which is the default under a converter specifies a different
+        one.
+        """
+
+        self._blocks = [None]
 
 
     def _add_converters(self):
@@ -571,35 +640,26 @@ class DiffConfig:
 
     def _add_converter(self, cvt):
         """Add an individual converter object, a child of the
-        DiffConverter class, to the list of converters to be used by the
-        convert() method.
+        DiffConverter class, to the list of converters for its block.
 
-        If the converter has an 'insert_before' entry that is not None,
-        the converter will be inserted in the postition immediately
-        before the existing entry with that name.  If that entry cannot
-        be found, a KeyError will be raised.
+        If the block used in the converter does not exist, a KeyError
+        will be raised.
         """
 
-        if not cvt.insert_before:
-            # no 'insert_before' - just add it to the end
-            self._cvts.append(cvt)
+        self._cvts[cvt.block].append(cvt)
 
-        else:
-            # 'insert_before' so we need to find that entry
 
-            # get the indexes and entries in the converter list
-            for pos, chk_cvt in enumerate(self._cvts):
-                # if this entry is the one we're inserting before, put
-                # it here and stop
-                if chk_cvts.name == cvts.insert_before:
-                    self._cvts.insert(pos, cvt)
-                    break
+    def _sort_converters(self):
+        """This method sorts the converters within each block.  The
+        default ordering is the the complete path (context + cmd + ext)
+        but can be overridden, if required.
 
-            else:
-                # we got the end of the list and didn't find it, so
-                # raise an exception
-                raise KeyError("cannot find existing converter with name: "
-                               + cvt.insert_before)
+        This method is called by the constructor, after the converters
+        have been added to ensure a consist order.
+        """
+
+        for block in self._cvts:
+            self._cvts[block].sort(key=lambda c: c._path_full)
 
 
     def _explain_comment(self, path):
@@ -856,6 +916,12 @@ class DiffConfig:
         diffs_tree = {}
 
 
+        # initialise the dictionary of activated triggers and their
+        # configuration points
+
+        triggers = {}
+
+
         # if no 'to' config was specified, stop here (we assume we're
         # just testing, parsing and excluding items from the 'from'
         # configuration and stopping
@@ -908,24 +974,42 @@ class DiffConfig:
                   sep="\n")
 
 
-        # go through the list of converter objects
+        # go through the list of blocks and then the converter objects
+        # within them, in order
+
+        for cvt in chain.from_iterable(
+            [ self._cvts[s] for s in self._blocks ]):
+
 
-        for cvt in self._cvts:
             # get all the remove and update matches for this converter
             # and combine them into one list, discarding any duplicates
             #
             # we do this rather than processing each list one after the
             # other so we combine removes and updates on the same part
             # of the configuration together
-            #
-            # we combine them by manually building a list, rather than
-            # using a set as a set requires hashable types
 
             remove_matches = cvt.full_matches(remove_cfg_full)
             update_matches = cvt.full_matches(update_cfg)
 
+
+            # if this converter is in a block which has been triggered,
+            # find all the matches in the 'to' configuration which match
+            # the context in which the trigger was set
+
+            triggered_matches = []
+            if cvt.block in triggers:
+                for t in triggers[cvt.block]:
+                    triggered_matches.extend(
+                        cvt.explicit_context_matches(to_cfg, t))
+
+
+            # combine all the matches for this converter into a single
+            # sorted list, removing duplicates
+
             all_matches = []
-            for match in sorted(remove_matches + update_matches):
+            for match in sorted(
+                remove_matches + update_matches + triggered_matches):
+
                 if match not in all_matches:
                     all_matches.append(match)
 
@@ -941,8 +1025,7 @@ class DiffConfig:
                 or (all_matches and self._debug_convert)):
 
                 print(">>", pathstr([ "*" if i is None else i
-                                    for i in cvt.get_path() ]),
-                      "name:", cvt.name)
+                                    for i in cvt.get_path() ]))
 
                 print()
 
@@ -1069,6 +1152,20 @@ class DiffConfig:
                         deepsetdefault(diffs_tree, *match)
 
 
+                        # if this converter sets off some triggers,
+                        # add those to the active trigger dictionary,
+                        # along with the matching patch
+
+                        for t in cvt.triggers:
+                            match_context = cvt.get_match_context(match)
+
+                            if self._debug_convert:
+                                print("=> trigger:",
+                                      pathstr(match_context, set()),
+                                      "block:", t)
+
+                            triggers.setdefault(t, []).append(match_context)
+
                     else:
                         if self._debug_convert:
                             print("-> no action")
@@ -1164,6 +1261,93 @@ class DiffConfig:
 
                         deepsetdefault(diffs_tree, *match)
 
+                        for t in cvt.triggers:
+                            match_context = cvt.get_match_context(match)
+
+                            if self._debug_convert:
+                                print("=> trigger:",
+                                      pathstr(match_context, set()),
+                                      "block:", t)
+
+                            triggers.setdefault(t, []).append(match_context)
+
+                    else:
+                        if self._debug_convert:
+                            print("-> no action")
+
+
+                # handle TRIGGERED conversions, if matching, and this
+                # wasn't already a remove or update conversion
+
+                if ((match in triggered_matches)
+                    and (match not in remove_matches)
+                    and (match not in update_matches)):
+
+                    if self._debug_convert:
+                        print("=> trigger match:",
+                              pathstr(match, cvt.wildcard_indices),
+                              "block:", cvt.block)
+
+
+                    args = cvt.get_wildcard_args(match)
+
+
+                    # get the 'from' and 'update' parts of the
+                    # configuration and update difference dictionaries,
+                    # for the path specified in the converter (ignoring
+                    # the extension 'ext')
+
+                    cvt_old = cvt.get_cfg(from_cfg, match)
+                    cvt_upd = cvt.get_cfg(update_cfg, match)
+                    cvt_new = cvt.get_cfg(to_cfg, match)
+
+                    if self._debug_convert >= 2:
+                        print("-> from configuration:",
+                              yaml.dump(cvt_old, default_flow_style=False),
+                              "-> update configuration:",
+                              yaml.dump(cvt_upd, default_flow_style=False),
+                              "-> to configuration:",
+                              yaml.dump(cvt_new, default_flow_style=False),
+                              sep="\n")
+
+
+                    # call the trigger converter action method
+
+                    try:
+                        diff = cvt.trigger(cvt_old, cvt_upd, cvt_new, *args)
+
+                    except:
+                        print("trigger() with:",
+                              "  type(cvt)=" + repr(type(cvt)),
+                              "  *args=" + repr(args),
+                              sep="\n", file=sys.stderr)
+
+                        raise
+
+
+                    # (same as in remove, above)
+
+                    if diff is not None:
+                        if isinstance(diff, str):
+                            diff = [diff]
+
+                        if self._debug_convert:
+                            print("\n".join(diff))
+
+
+                        if self._explain:
+                            comment = self._explain_comment(match)
+                            if comment:
+                                diffs.append(comment)
+
+
+                        diffs.extend(diff)
+                        diffs.append("")
+
+                        deepsetdefault(diffs_tree, *match)
+
+                        # except one trigger cannot fire another
+
                     else:
                         if self._debug_convert:
                             print("-> no action")
