@@ -476,11 +476,27 @@ class Cmd_Int_SwPortTrkNtv(CmdContext_Int):
 
 
 class Cmd_Int_SwPortTrkAlw(CmdContext_Int):
-    match = r"switchport trunk allowed vlan (add )?(?P<vlans>[0-9,-]+)"
+    match = (r"switchport trunk allowed vlan "
+             r"((?P<complete>(none|all))|(?P<add>add )?(?P<vlans>[0-9,-]+))")
+
+    def parse(self, cfg, complete, add, vlans):
+        if complete:
+            # 'all' is the same as no configuration present
+            if complete == "all":
+                if "switchport-trunk-allow" in cfg:
+                    cfg.pop("switchport-trunk-allow")
+
+            # 'none' is explicitly no VLANs
+            elif complete == "none":
+                cfg["switchport-trunk-allow"] = set()
+
+        elif add:
+            cfg.setdefault("switchport-trunk-allow", set()).update(
+                expand_set(vlans))
+
+        else:
+            cfg["switchport-trunk-allow"] = expand_set(vlans)
 
-    def parse(self, cfg, vlans):
-        cfg.setdefault("switchport-trunk-allow", set()).update(
-            expand_set(vlans))
 
 
 # ...
