@@ -480,7 +480,7 @@ class Cvt_Int_ServPol(CvtContext_Int):
     def type_str(self, type_):
         # we only specify the 'type' if one is set (the default is a
         # policer)
-        return " type %s" if type_ else ""
+        return (" type " + type_) if type_ else ""
 
     def remove(self, old, int_name, dir_, type_):
         return [*self.enter(int_name),
@@ -680,6 +680,16 @@ class Cvt_Int_SwPortTrkAlw(CvtContext_Int):
     cmd = tuple()
     ext = "switchport-trunk-allow",
 
+    def remove(self, old, int_name):
+        # if this interface is in a port-channel, we do all changes
+        # there, so skip this
+        if "channel-group" in old:
+            return None
+
+        # we're removing all commands allowing VLANs which is a special
+        # case as this actually means 'allow all'
+        return *self.enter(int_name), " no switchport trunk allowed vlan"
+
     def truncate(self, old, rem, int_name):
         # if this interface is in a port-channel, we do all changes
         # there, so skip this
@@ -698,8 +708,16 @@ class Cvt_Int_SwPortTrkAlw(CvtContext_Int):
             return None
 
         l = [*self.enter(int_name)]
+
+        # if this list was not there before, all VLANs were allowed by
+        # default so we need to reset the list to 'none' and then add
+        # the ones which are specifically listed
+        if not old:
+            l.append(" switchport trunk allowed vlan none")
+
         for tag in sorted(self.get_ext(upd)):
             l.append(" switchport trunk allowed vlan add " + str(tag))
+
         return l
 
 
